package Language::Basic::Expression;
# Part of Language::Basic by Amir Karger (See Basic.pm for details)

=pod

=head1 DESCRIPTION

Language::Basic::Expression and its subclasses handle string, numeric, and
conditional expressions. 

Each subclass has (at least) two methods: parse takes a string ref and digests
it (or returns undef if the string's syntax doesn't match that subclass);
evaluate actually calculates the value of the expression. For a string
or numeric constant or variable, that just means taking the stored value
of that object. For other Expressions, you actually need to do math.

The Expression subclasses closely follow the BASIC grammar. Most subclasses
can have their own String and Numeric subclasses in turn. This allows us
to make sure that you don't try to multiply strings together, etc.

=cut

use strict;
use Language::Basic::Common;

# sub-packages
{
package Language::Basic::Expression::Conditional;
package Language::Basic::Expression::Arithmetic;
package Language::Basic::Expression::Multiplicative;
package Language::Basic::Expression::Unary;
package Language::Basic::Expression::Lvalue;
package Language::Basic::Expression::Function;
package Language::Basic::Expression::Constant;
}

# new blesses an empty object to the class you're in, then parses
# the text (whose reference is given in arg1) to populate the object
# with different fields, depending on the class.
# The parse subroutine may well call other 'new' subs and return a whole
# hierarchy of LB::Expressions. Note that this sub is inherited by
# all the LB::Expression::* subclasses.
sub new {
    # important to use shift here, since parse passes any remaining @_
    my ($class, $textref) = (shift, shift);
    my $self = {};
    bless $self, $class;
    # parse the expression, returns subclass (or undef for a problem)
    # Also pass any extra arguments that new got
    $self->parse($textref, @_) or return undef; 
    return $self;
} # end sub Language::Basic::Expression::new

# An arithmetic expression is a LBE::Arithmetic::String if it's made up
# of LBE::Multiplicative::String expressions and so on. We never mix
# expression types (except within an argument list for an array or function.)
# This sub therefore blesses an object to its String or Numeric subclass
# depending on the type of the sub-expression (and returns the newly
# blessed object.)
# Arg0 is the thing to bless, arg1 is the subexp
sub string_or_numeric {
    my $self = shift;
    my $class = ref($self);
    # If we already are blessed, don't rebless!
    return $self if ref($self) =~ /::(String|Numeric)$/;

    my $subexp = shift;
    ref($subexp) =~ /::(String|Numeric)$/ or 
        die "Error refining $class to ",ref($subexp),"\n";
    my $type = $1;
    #print "self, class, type is 1 $self 2 $class 3 $type\n";
    # Note: "$class::$type" breaks!
    my $subclass = $class . "::$type";
    bless $self, $subclass;
}

######################################################################
{
# An arithmetic expression is a set of multiplicative expressions connected by
# + or - (String expressions can only be connected by +)

package Language::Basic::Expression::Arithmetic;
@Language::Basic::Expression::Arithmetic::ISA = qw(Language::Basic::Expression);

sub parse {
# The while loop is necessary in case we have an expression like 1+2+3 
# It will effectively evaluate the +, - operators left to right 
    my $self = shift;
    my $textref = shift;
    my (@adds, @ops);

    my $add = new Language::Basic::Expression::Multiplicative $textref;
    push @adds, $add;
    while ($$textref =~ s/^([+-])//) {
	push @ops, $1;
	$add = new Language::Basic::Expression::Multiplicative $textref;
	push @adds, $add;
    } # end while

    $self->{"expressions"} = \@adds;
    $self->{"operations"} = \@ops;
    # Bless to LBEA::String or Numeric
    $self->string_or_numeric($add);
} # end sub Language::Basic::Expression::Arithmetic::parse

package Language::Basic::Expression::Arithmetic::String;
@Language::Basic::Expression::Arithmetic::String::ISA = qw(Language::Basic::Expression::Arithmetic);

sub evaluate {
    my $self = shift;
    my @exps = @{$self->{"expressions"}};
    # Ops ought to be all pluses, since that's all BASIC can do.
    my @ops = @{$self->{"operations"}};

    my $add = (shift @exps)->evaluate;
    while (my $op = shift @ops) {
	my $add2 = (shift @exps)->evaluate;
	if ($op eq '+') {
	    $add .= $add2;
	} else { # minus
	    die "Unknown op in LBE::Arithmetic::String::evaluate!\n";
	}
    } # end while
    return($add);
} # end sub Language::Basic::Expression::Arithmetic::evaluate

package Language::Basic::Expression::Arithmetic::Numeric;
@Language::Basic::Expression::Arithmetic::Numeric::ISA = qw(Language::Basic::Expression::Arithmetic);

sub evaluate {
    my $self = shift;
    my @exps = @{$self->{"expressions"}};
    my @ops = @{$self->{"operations"}};

    my $add = (shift @exps)->evaluate;
    while (my $op = shift @ops) {
	my $add2 = (shift @exps)->evaluate;
	#print "eval_add: $add $op $add2\n";
	if ($op eq '+') {
	    $add = $add + $add2;
	} else { # minus
	    $add = $add - $add2;
	}
    } # end while
    return($add);
} # end sub Language::Basic::Expression::Arithmetic::evaluate

} # end package Language::Basic::Expression::Arithmetic

{
# A multiplicative expression is a set of unary expressions connected by * or / 
# A String multiplicative expression is just a String unary expression.
package Language::Basic::Expression::Multiplicative;
@Language::Basic::Expression::Multiplicative::ISA = qw(Language::Basic::Expression);

sub parse {
# Note that String unary expressions can't multiply or divide, but it doesn't
# really matter
    my $self = shift;
    my $textref = shift;
    my (@mults, @ops);

    my $mult = new Language::Basic::Expression::Unary $textref;
    push @mults, $mult;
    while ($$textref =~ s/^([*\/])//) {
	push @ops, $1;
	$mult = new Language::Basic::Expression::Unary $textref;
	push @mults, $mult;
    } # end while

    $self->{"expressions"} = \@mults;
    $self->{"operations"} = \@ops;

    # Bless to LBEM::String or Numeric
    $self->string_or_numeric($mult);
} # end sub Language::Basic::Expression::Multiplicative::parse

sub evaluate {
    my $self = shift;
    my @exps = @{$self->{"expressions"}};
    my @ops = @{$self->{"operations"}};

    my $mult = (shift @exps)->evaluate;
    while (my $op = shift @ops) {
	my $mult2 = (shift @exps)->evaluate;
	#print "eval_mult: $mult $op $mult2\n";
	if ($op eq '*') {
	    $mult = $mult * $mult2;
	} else {
	    $mult = $mult / $mult2;
	}
    } # end while
    return($mult);
} # end sub Language::Basic::Expression::Multiplicative::evaluate

# Sub packages
package Language::Basic::Expression::Multiplicative::Numeric;
@Language::Basic::Expression::Multiplicative::Numeric::ISA = qw(Language::Basic::Expression::Multiplicative);
package Language::Basic::Expression::Multiplicative::String;
@Language::Basic::Expression::Multiplicative::String::ISA = qw(Language::Basic::Expression::Multiplicative);
} # end package Language::Basic::Expression::Multiplicative


{
package Language::Basic::Expression::Unary;
@Language::Basic::Expression::Unary::ISA = qw(Language::Basic::Expression);
use Language::Basic::Common;

sub parse {
# This function parses a unary expression. That is, a variable,
# a string or numeric constant, or an arithmetic expression in parentheses,
# potentially with a unary minus sign.
    my $self = shift;
    my $textref = shift;

    my $unary;
    my $try;

    $self->{"minus"} = ($$textref =~ s/^-//); # unary minus in the expression?

    # if a parentheses, call eval_arithmetic recursively on what's inside
    if ($$textref =~ s/^\(//)  {
	$unary = new Language::Basic::Expression::Arithmetic $textref;
	$$textref =~ s/^\)// or warn "missing ')' character?";  # skip end paren

    # OR it's a function
    # NOTE that LBEF::new had better not eat the word if it returns undef!
    } elsif (defined ($try = 
	    new Language::Basic::Expression::Function $textref)) {
	$unary = $try;

    # OR it's a String or Numeric variable
    } elsif (defined ($try = 
	    new Language::Basic::Expression::Lvalue $textref)) {
	$unary = $try;

    # OR it's a String or Numeric constant
    } elsif (defined ($try = 
	    new Language::Basic::Expression::Constant $textref)) {
	$unary = $try;

    # Or die
    } else {
        Exit_Error("Unknown expression $$textref");
    }

    $self->{"expression"} = $unary;
    #print "unary ref is ",ref($unary),"\n";

    # Bless to LBEU::String or Numeric
    $self->string_or_numeric($unary);
} # end eval_unary

sub evaluate {
# This function evaluates a unary expression. That is, a variable,
# a number, or an arithmetic expression in parentheses,
# potentially with a unary minus sign.
# 
    my $self = shift;
    my $exp = $self->{"expression"};

    my $value = $exp->evaluate;
    $value = -$value if $self->{"minus"};
    return($value);
} # end sub Language::Basic::Expression::Unary::evaluate

# Sub packages
package Language::Basic::Expression::Unary::Numeric;
@Language::Basic::Expression::Unary::Numeric::ISA = qw(Language::Basic::Expression::Unary);
package Language::Basic::Expression::Unary::String;
@Language::Basic::Expression::Unary::String::ISA = qw(Language::Basic::Expression::Unary);
} # end package Language::Basic::Expression::Unary

######################################################################
{
# A string or numeric constant
package Language::Basic::Expression::Constant;
@Language::Basic::Expression::Constant::ISA = qw(Language::Basic::Expression);

# Returns a LBE::Constant::* subclass or undef
# TODO this totally sucks. Because of the complicatedness of the new/parse
# scheme, we have to cannibalize the returned LBEC::* object and
# call string_or_numeric, rather than just returning that object.
# Maybe LBE::new should say "$self = $self->parse". That would require
# having all the other parse subs return self, but that's doable.
sub parse {
    my $self = shift;
    my $textref = shift;
    my ($const, $try);
    if (defined ($try = 
	    new Language::Basic::Expression::Constant::Numeric $textref)) {
	$const = $try;
    } elsif (defined ($try = 
	    new Language::Basic::Expression::Constant::String $textref)) {
	$const = $try;
    } else {
        return undef;
    }

    $self->{"value"} = $const->{"value"};
    $self->string_or_numeric($const);
} # end Language::Basic::Expression::Constant::parse

sub evaluate {return shift->{"expression"}->evaluate; }

package Language::Basic::Expression::Constant::Numeric;
@Language::Basic::Expression::Constant::Numeric::ISA = qw(Language::Basic::Expression::Constant);

sub parse {
    my $self = shift;
    my $textref = shift;
    if ($$textref =~ s/^(\d*\.?\d+)//) {
	$self->{"value"} = $1 + 0;
    } else {
        return undef;
    }
    # Otherwise, if value is empty string, returning it confuses LBE::new
    return $self;
} # end sub Language::Basic::Expression::Constant::Numeric::parse

sub evaluate { return shift->{"value"} }

package Language::Basic::Expression::Constant::String;
@Language::Basic::Expression::Constant::String::ISA = qw(Language::Basic::Expression::Constant);

sub parse {
    my $self = shift;
    my $textref = shift;
    if ($$textref =~ s/^"(.*?)"//) {
	$self->{"value"} = $1;
    } else {
	# TODO handle unquoted string for Input, Data statements
	warn "Currently only understand quoted strings for String Constant: $$textref";
        return undef;
    }
    # Otherwise, if value is empty string, returning it confuses LBE::new
    return $self;
} # end sub Language::Basic::Expression::Constant::String::parse

sub evaluate { return shift->{"value"} }

} # end package Language::Basic::Expression::Constant

######################################################################
# package Language::Basic::Expression::Lvalue
# An lvalue is a settable expression: a variable or one cell in an array.
#
# Fields:
#    varptr - ref to the LB::Variable (::Array or ::Scalar) object. Note
#        that it does NOT ref a particular cell in an LBV::Array object!
#    arglist - a set of Arithmetic Expressions describing which exact cell
#        in an LBV::Array to get. undef for a LBV::Scalar
{
package Language::Basic::Expression::Lvalue;
@Language::Basic::Expression::Lvalue::ISA = qw(Language::Basic::Expression);
use Language::Basic::Common;

# Sub-packages
{
package Language::Basic::Expression::Lvalue::Numeric;
@Language::Basic::Expression::Lvalue::Numeric::ISA = qw(Language::Basic::Expression::Lvalue);
package Language::Basic::Expression::Lvalue::String;
@Language::Basic::Expression::Lvalue::String::ISA = qw(Language::Basic::Expression::Lvalue);
}

sub parse {
    my $self = shift;
    my $textref = shift;
    $$textref =~ s/^[A-Z]\w*\$?// or return undef;
    my $name = $&;

    # read ( Arglist ) if it exists
    # By default, though, it's a scalar, and has no ()
    $self->{"arglist"} = undef;
    if (defined (my $arglist = 
	    new Language::Basic::Expression::Arglist $textref)) {
	$self->{"arglist"} = $arglist;
    }

    # Look up the variable by name in the (Array or Scalar) variable storage.
    # (Also, create the Variable if it doesn't yet exist.)
    my $var = &Language::Basic::Variable::lookup($name, $self->{"arglist"});
    $self->{"varptr"} = $var;

    # Is it a string or numeric lvalue?
    $self->string_or_numeric($var);
} # end sub Language::Basic::Expression::Lvalue::parse

sub evaluate { 
    my $self = shift;
    # This automatically gets the correct array cell if necessary
    my $var = $self->variable;
    my $value = $var->value;
    return $value;
} # end sub Language::Basic::Expression::Lvalue::evaluate

# returns a variable, e.g. for setting in a Let or changing in a Next
# Note that it always returns a LB::Variable::Scalar object. If the
# variable in this expression is an Array, it returns one cell from the array.
sub variable { 
    my $self = shift;
    my $var = $self->{"varptr"};
    # if Arglist exists, evaluate each arith. exp. in it and get that cell
    # from the Array
    if (defined (my $arglist = $self->{"arglist"})) {
	my @args = $arglist->evaluate;
	$var = $var->get_cell(@args);
    }

    return $var;
} # end sub Language::Basic::Expression::Lvalue::variable

} # end package Language::Basic::Expression::Lvalue

######################################################################
# package Language::Basic::Expression::Function
#
# Fields:
#    function - ref to the LB::Function (::Intrinsic or ::Defined) used
#        by this expression
#    arglist - a set of Arithmetic Expressions describing the arguments
#        to pass to the function
{
package Language::Basic::Expression::Function;
@Language::Basic::Expression::Function::ISA = qw(Language::Basic::Expression);
use Language::Basic::Common;

# Sub-packages
{
package Language::Basic::Expression::Function::Numeric;
@Language::Basic::Expression::Function::Numeric::ISA = qw(Language::Basic::Expression::Function);
package Language::Basic::Expression::Function::String;
@Language::Basic::Expression::Function::String::ISA = qw(Language::Basic::Expression::Function);
}

# Arg0, Arg1 are the object and a ref to the string being parsed, as usual.
# Arg2, if it exists, says we're in a DEF statement, so that if the
# function doesn't exist, we should create it rather than returning undef.
sub parse {
    my $self = shift;
    my $textref = shift;
    # Note: we can't eat the func name if it turns out not to be a function!
    $$textref =~ /^[A-Z]\w*\$?/ or return undef;
    my $name = $&;
    my $defining = (defined (my $exp = shift));

    # Look up the function name
    # If the function doesn't exist, the word is a variable or something...
    # Alternatively, if there was a second argument to parse, then we're
    # in a DEF statement & should create the function.
    my $func;
    if ($defining) {
        $func = new Language::Basic::Function::Defined $name;
    } else {
	$func = &Language::Basic::Function::lookup($name) or return undef;
    }

    # Now that we know it's a function, eat the name
    $$textref =~ s/^[A-Z]\w*\$?//;
    $self->{"function"} = $func;

    # read ( Arglist )
    if ($defining) {
	$$textref =~ s/\((.*)\)// or 
	    Exit_Error("DEF missing function arguments!");
	my $argtext = $1;
	my @vars;
	do {
	    my $arg = new Language::Basic::Expression::Lvalue \$argtext;
	    # TODO test that arg is a Scalar!
	    push @vars, $arg->variable;
	} while $argtext =~ s/^,//;

        # Define the number & type of args in the subroutine, as well as the
	# expression that defines the function.
	$func->define (\@vars, $exp);

    } else {
	my $arglist = new Language::Basic::Expression::Arglist $textref
	    or Exit_Error("Function without arglist!");
	# check if the number or type of args is wrong.
	my $error = $func->check_args($arglist);
	Exit_Error($error) if $error;
	$self->{"arglist"} = $arglist;
    }

    # Is it a string or numeric Function?
    $self->string_or_numeric($func);
} # end sub Language::Basic::Expression::Function::parse

sub evaluate { 
    my $self = shift;
    my $func = $self->{"function"};
    my $arglist = $self->{"arglist"};
    # Note we tested number & type of args in parse
    my @args = $arglist->evaluate;
    my $value = $func->evaluate(@args);
    return $value;
} # end sub Language::Basic::Expression::Function::evaluate

} # end package Language::Basic::Expression::Function

{
# A list of arguments to an array/function
package Language::Basic::Expression::Arglist;
@Language::Basic::Expression::Arglist::ISA = qw(Language::Basic::Expression);
use Language::Basic::Common;

sub parse {
    my $self = shift;
    my $textref = shift;

    # Has to start with paren
    $$textref =~ s/^\(// or return undef;

    # Eat args
    my @args = ();
    do {
        my $arg = new Language::Basic::Expression::Arithmetic $textref;
	push @args, $arg;
    } while $$textref =~ s/^,//;

    # Has to end with paren
    $$textref =~ s/^\)// or Exit_Error("Arglist without ')' at end!");
    $self->{"arguments"} = \@args;
} # end sub Language::Basic::Expression::Arglist::parse

# Returns a LIST of values
sub evaluate {
    my $self = shift;
    my @values = map {$_->evaluate} @{$self->{"arguments"}};
    return @values;
} # end sub Language::Basic::Expression::Arglist::evaluate

} # end package Language::Basic::Expression::Arglist

######################################################################
# package Language::Basic::Expression::Conditional
# A BASIC string expression (e.g., A$, or "hello there")
{
package Language::Basic::Expression::Conditional;
@Language::Basic::Expression::Conditional::ISA = qw(Language::Basic::Expression);

sub parse {
    my ($self, $textref) = @_;
    my $e1 = new Language::Basic::Expression::Arithmetic $textref;

    $$textref =~ s/^[<=>]+// or Exit_Error("No Rel. Op. in Conditional Exp.");
    my $op = $&;

    my $e2 = new Language::Basic::Expression::Arithmetic $textref;
    $self->{"exp1"} = $e1;
    $self->{"exp2"} = $e2;

    # Convert BASIC ops to perlops
    my $num_op = {
        "="  => "==",
	">"  => ">",
	"<"  => "<",
	">=" => ">=",
	"<=" => "<=",
	"<>" => "!=",
    };
    my $string_op = {
        "="  => "eq",
	">"  => "gt",
	"<"  => "lt",
	">=" => "ge",
	"<=" => "le",
	"<>" => "ne",
    };
    my $trans = (ref($e1) =~ /String/ ? $string_op : $num_op);
    my $perlop = $trans->{$op} or Exit_Error("Unrecognized Rel. op. '$op'");
    $self->{"op"} = $perlop;
    
    # Is conditional String or Numeric? Figure out from its exps
    Exit_Error("Expressions in conditional must be the same type!") unless
        ref($e1) eq ref($e2);
    $self->string_or_numeric($e1);
} # end sub Language::Basic::Expression::new

sub evaluate {
    my $self = shift;

    # Evaluate the two sides of the conditional (each is an expression--
    # either they're both arithmetic or they're both string).
    my $exp1 = $self->{"exp1"};
    my $exp2 = $self->{"exp2"};
    my $e1 = $exp1->evaluate;
    my $e2 = $exp2->evaluate;

    my $perlop = $self->{"op"};
    # I'm vainly hoping that Perl eval will get the same result BASIC would
    # Need to use \Q in case we say IF A$ = "\", which should really compare
    # with \\.
    my $perlexp = join ("", "\"\Q$e1\E\" ", $perlop . " \"\Q$e2\E\"");
    my $value = eval $perlexp;
    #print "exp is '$perlexp', value is '$value'\n";
    
    return $value;
} # end sub Language::Basic::Expression::Conditional::evaluate

# Sub packages
package Language::Basic::Expression::Conditional::Numeric;
@Language::Basic::Expression::Conditional::Numeric::ISA = qw(Language::Basic::Expression::Conditional);
package Language::Basic::Expression::Conditional::String;
@Language::Basic::Expression::Conditional::String::ISA = qw(Language::Basic::Expression::Conditional);
} # end package Language::Basic::Expression::Conditional

1; # end package Language::Basic::Expression
